#define CATCH_CONFIG_MAIN  

#include "catch.hpp"

extern "C" {
	#include "datashadow.h"
}

TEST_CASE("Test edge/error cases", "[datashadow]" ) {

	// this setup code is shared across sections w/in the test case	
	initialize_data_shadow_module();

	SECTION("Test out-of-bounds shadow access") {
		REQUIRE ( add_shadow_entry(-5, 0L)==0 );
		REQUIRE ( add_shadow_entry(5000000, 0L)==0 );
	}

	SECTION("Test non-existent entry") {
		void *restored = NULL;
		int count = 0;
		REQUIRE ( check_shadow_value(5, (void*)0x12345678L, &restored, &count) == 0 );
	}
}

TEST_CASE( "Test normal operations", "[datashadow]" ) {

	void *restored = NULL;
	int count = 0;

	initialize_data_shadow_module();

	SECTION( "Test single shadow entry") {
		void *address = (void*) 0x12345678L;

		REQUIRE ( add_shadow_entry(0, address) == 1 );

		REQUIRE ( check_shadow_value(0, address, &restored, &count) == 1 );
		REQUIRE ( restored == address );
	}

	SECTION( "Test multiple shadow entries") {
		void *address1 = (void*) 0x12345678L;
		void *address2 = (void*) 0x11111111L;
		void *address3 = (void*) 0x22222222L;
		void *bogus_address = (void*) 0xaabbccdd;

		add_shadow_entry(2, address1);
		add_shadow_entry(2, address2);
		add_shadow_entry(2, address3);

		REQUIRE ( check_shadow_value(2, address1, &restored, &count) == 1 );
		REQUIRE ( restored == address1 );

		REQUIRE ( check_shadow_value(2, address2, &restored, &count) == 1 );
		REQUIRE ( restored == address2 );

		REQUIRE ( check_shadow_value(2, address3, &restored, &count) == 1 );
		REQUIRE ( restored == address3 );

		REQUIRE ( check_shadow_value(2, bogus_address, &restored, &count) == 0);
		REQUIRE ( (restored == address1  || restored == address2 || restored == address3) );
	}
}
